/* Program demonstrira kreiranje niti. Nitima procesor na razlicite
   nacine dodeljuje vreme, pa je redosled izvrsavanja niti nepredvidiv
   i razlicit u raznim pokretanjima programa. */
#include <stdio.h>
#include <pthread.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>

#define NUM_THREADS 20

void error_fatal(char *message)
{
  perror(message);
  exit (EXIT_FAILURE);
}

/* Posebna funkcija koja se poziva kada dodje do greske pri radu sa
   nitima. */
void pthread_error(char *message, int err)
{
  /* Funkcija strerror na osnovu vrednosti celobrojnog koda greske
     vraca pokazivac na string koji detaljnije opisuje gresku. */
  fprintf(stderr, "%s : %s\n", message, strerror(err));
  exit (EXIT_FAILURE);
}

void* fun(void *arg)
{
  pthread_t tid = pthread_self();
  /* Najcesce je pthread_t implementiran da bude unsigned int, pa
     radimo kastovanje i pokusavamo da ispisemo thread id. Ovo nije
     portabilan nacin da se ispise thread id, ali ce na vecini
     implementacija raditi. */
  printf("thread index: %-3d,\ttid: %-8u,\tpid: %d\n", *((int*)arg), (unsigned int)tid, getpid());
  free (arg);
  return ((void*)0);
}

int main()
{
  int err;
  int i;
  pthread_t tid;
  for(i = 0; i < NUM_THREADS; i++)
  {
    /* Neophodno je koristiti dinamicku alokaciju jer funkcije za rad
       sa nitima primaju pokazivace. Sta se desava ukoliko se
       alokacija prebaci pre for petlje? Razlicite niti onda
       pristupaju istoj memoriji, sto nije namera - pokrenuti program
       na ovaj nacin i uveriti se u ovaj scenario. */
    int *num = (int*)malloc(sizeof(int));
    if(num == NULL)
      error_fatal ("malloc error");
    *num = i;
    /* int k = i; -- ne savetuje se koriscenje ovog nacina bez
       dinamicke alokacije (u tom slucaju se u pozivu pthread_create
       prosledjuje &k, a eliminise dealokacija memorije u okviru
       niti). Ovakav nacin ce nekad raditi, nekad ne, zavisno od tipa
       podataka koji se koristi. */

    /* Kreiranje niti. Do greske moze doci ako se na primer kreira
       preveliki broj niti (probati sa hiljadama niti. Kod niti se
       status greske cuva u povratnoj vrednosti, ne postavlja se
       promenljiva errno, pa se koristi posebna funkcija za ispis
       greske. */
    if( (err = pthread_create(&tid, NULL, fun, num)) != 0)
      pthread_error ("create thread error", err);
  }

  /* Glavni program ceka da se sve niti zavrse. U suprotnom bi moglo
     da se desi da se main završi pre ostalih niti, pa bi mogao ceo
     proces da se završi pre nego što su neke niti i
     pokrenute. Koriscenje sleep funkcije nije dobar nacin da se ovo
     uradi, pogledati naredni primer. */
  sleep(3);
  printf ("The end of main function, pid : %d.\n", getpid());
  
  return 0;
}
