/* Program demonstrira cekanje na zavrsetak niti. Zato sto glavna nit
   (program) ceka na svaku nit da se zavrsi, niti ce sigurno biti
   zavrsene redosledom kojim su kreirane. Ovo nema smisla, jer ideja u
   radu sa nitima je upravo da se izvrsavaju paralelno a ne
   sekvencijalno.*/
#include <stdio.h>
#include <pthread.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>

#define NUM_THREADS 20

void error_fatal(char *message)
{
  perror(message);
  exit (EXIT_FAILURE);
}

/* Posebna funkcija koja se poziva kada dodje do greske pri radu sa
   nitima. */
void pthread_error(char *message, int err)
{
  /* Funkcija strerror na osnovu vrednosti celobrojnog koda greske
     vraca pokazivac na string koji detaljnije opisuje gresku. */
  fprintf(stderr, "%s : %s\n", message, strerror(err));
  exit (EXIT_FAILURE);
}

void* fun(void *arg)
{
  *((int*)arg) = -(*((int*)arg));
  return arg;
}

int main()
{
  int err;
  int i;
  void *ret_value;

  /* Alokacija u ovom slucaju je mogla biti i staticka posto je broj
     niti unapred poznat. */
  pthread_t* tid_a = malloc(NUM_THREADS * sizeof(pthread_t));
  int *thread_arguments = malloc(NUM_THREADS * sizeof(int));
  if(tid_a == NULL || thread_arguments == NULL)
    error_fatal ("malloc error");

  /* Kreiranje niti. */
  for(i = 0; i < NUM_THREADS; i++)
  {
    thread_arguments[i] = i;

    /* Kreiranje niti. */
    if( (err = pthread_create(&tid_a[i], NULL, fun, &thread_arguments[i])) != 0)
      pthread_error ("thread create error", err);
  }

  /* Cekanje na niti tek kada su sve kreirane. */
  for(i = 0; i < NUM_THREADS; i++)
  {
    if( (err = pthread_join(tid_a[i], &ret_value)) != 0)
      pthread_error ("pthread join error", err);
    printf("%d\n", *((int*)ret_value));
  }

  /* Oslobadjanje dinamicki alocirane memorije. */
  free (tid_a);
  free (thread_arguments);

  return 0;
}
