/* Program demonstrira nepredvidivost ispisa kada nema sinhronizacije niti. 
   Sledece tri operacije ne predstavljaju jednu atomicnu operaciju:
   1. Učitaj vrednost iz memorije u registar,
   2. Izmeni vrednost u registru,
   3. Upiši novu vrednost nazad u memoriju.
   pa posto procesor nepredvidivim redosledom dodeljuje nitima procesorsko vreme,
   vrednost izlazne promenljive vrednosti x moze imati razlicite vrednosti u
   raznim pokretanjima programa. */
#include <stdio.h>
#include <pthread.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>

#define NUM_THREADS 200
#define NUM_LOOPS 1000

/* Globalna promenljiva koju menjaju sve niti. */
int x;

void error_fatal(char *message)
{
  perror(message);
  exit (EXIT_FAILURE);
}

/* Posebna funkcija koja se poziva kada dodje do greske pri radu sa nitima. */
void pthread_error(char *message, int err)
{
  /* Funkcija strerror na osnovu vrednosti celobrojnog koda greske vraca pokazivac
     na string koji detaljnije opisuje gresku. */
  fprintf(stderr, "%s : %s\n", message, strerror(err));
  exit (EXIT_FAILURE);
}

/* Funkcija uvecava vrednost promenljive x NUM_LOOPS puta za po 1. */
void* fun1(void *arg)
{
  int i;
  for(i = 0; i < NUM_LOOPS; i++)
    x+=1;
  return ((void*)0);
}

/* Funkcija umanjuje vrednost promenljive x NUM_LOOPS puta za po 1. */
void* fun2(void *arg)
{
  int i;
  for(i = 0; i < NUM_LOOPS; i++)
    x-=1;
  return ((void*)0);
}

int main()
{
  int err;
  int i;
  /* Alokacija u ovom slucaju je mogla biti i staticka posto je broj
     niti unapred poznat. */
  pthread_t* tid_a = malloc(NUM_THREADS * sizeof(pthread_t));
  if(tid_a == NULL)
    error_fatal ("malloc error");

  /* Kreiranje niti. */
  for(i = 0; i < NUM_THREADS/2; i++)
  {
    /* Polovina niti ce pozivati funkciju fun1, a druga polovina funkciju fun2. */
    if( (err = pthread_create(&tid_a[i], NULL, fun1, NULL)) != 0)
      pthread_error ("thread create error", err);
    if( (err = pthread_create(&tid_a[NUM_THREADS/2 + i], NULL, fun2, NULL)) != 0)
      pthread_error ("thread create error", err);
  }

  /* Cekanje na sve niti: ovo je pravilan nacin. */
  for(i = 0; i < NUM_THREADS; i++)
  {
    if( (err = pthread_join(tid_a[i], NULL)) != 0)
      pthread_error ("thread join error", err);
  }

  /* Stampanje vrednosti x.v i oslobadjanje dinamicki alocirane memorije. */
  printf("%d\n", x);
  free(tid_a);
  return 0;
}
