/* Program demonstrira koriscenje mutexa. Promenljiva se zakljucava u
   delu koda u kome se uvecava njena vrednost pa ostale niti ne mogu
   da menjaju vrednost promenljive dok ona ne bude
   otkljucana. Rezultat programa ce biti isti u svim pokretanjima. */
#include <stdio.h>
#include <pthread.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>

#define NUM_THREADS 300
#define NUM_LOOPS 1000

/* Deklarisemo posebnu strukturu koja cuva promenljivu i njen
   muteks. */
typedef struct _sync_int
{
  int v;
  pthread_mutex_t lock;
}sync_int;

sync_int x;

void error_fatal(char *message)
{
  perror(message);
  exit (EXIT_FAILURE);
}

/* Posebna funkcija koja se poziva kada dodje do greske pri radu sa
   nitima. */
void pthread_error(char *message, int err)
{
  /* Funkcija strerror na osnovu vrednosti celobrojnog koda greske
     vraca pokazivac na string koji detaljnije opisuje gresku. */
  fprintf(stderr, "%s : %s\n", message, strerror(err));
  exit (EXIT_FAILURE);
}

/* Funkcija uvecava vrednost promenljive x NUM_LOOPS puta za po 1. */
void* fun(void *arg)
{
  int i, err;
  for(i = 0; i < NUM_LOOPS; i++)
  {
    /* Zakljucavanje varijable x za promene. */
    if ( (err = pthread_mutex_lock(&x.lock)) != 0)
      pthread_error ("mutex lock error", err);
    x.v += 1;
    /* Otkljucavanje varijable x. */
    if ( (err = pthread_mutex_unlock(&x.lock)) != 0)
      pthread_error ("mutex unlock error", err);
  }
  return ((void*)0);
}

/* Funkcija umanjuje vrednost promenljive x NUM_LOOPS puta za po 1. */
void* fun2(void *arg)
{
  int i, err;
  for(i = 0; i < NUM_LOOPS; i++)
  {
    /* Zakljucavanje varijable x za promene. */
    if ( (err = pthread_mutex_lock(&x.lock)) != 0)
      pthread_error ("mutex lock error", err);
    x.v -= 1;
    /* Otkljucavanje varijable x. */
    if ( (err = pthread_mutex_unlock(&x.lock)) != 0)
      pthread_error ("mutex unlock error", err);
  }
  return ((void*)0);
}

int main()
{
  int err;
  int i;
  x.v = 0;
  
  /* Neophodna je inicijalizacija muteksa pre nego sto ga je moguce koristiti. */
  if( (err = pthread_mutex_init(&x.lock, NULL)) != 0)
    pthread_error ("mutex init error", err);
  
  pthread_t* tid_a = malloc(NUM_THREADS * sizeof(pthread_t));
  if(tid_a == NULL)
    error_fatal ("malloc error");

  /* Kreiranje niti. */
  for(i = 0; i < NUM_THREADS/2; i++)
  {
    /* Polovina niti ce pozivati funkciju fun, a druga polovina funkciju fun2. */
    if( (err = pthread_create(&tid_a[i], NULL, fun, NULL)) != 0)
      pthread_error ("thread create error", err);
    if( (err = pthread_create(&tid_a[NUM_THREADS/2 + i], NULL, fun2, NULL)) != 0)
      pthread_error ("thread create error", err);
  }

  /* Cekanje na sve niti. */
  for(i = 0; i < NUM_THREADS; i++)
  {
    if( (err = pthread_join(tid_a[i], NULL)) != 0)
      pthread_error ("thread join error", err);
  }

  /* Neophodno brisanje muteksa na kraju. */
  if( (err = pthread_mutex_destroy(&x.lock)) != 0)
    pthread_error ("destroy mutex error", err);

  /* Stampanje vrednosti x.v i oslobadjanje dinamicki alocirane memorije. */
  printf("%d\n", x.v);
  free (tid_a);
  return 0;
}
