#include <stdio.h>
#include <pthread.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>

#define NUM_THREADS 2

void error_fatal(char *message)
{
  perror(message);
  exit (EXIT_FAILURE);
}

/* Posebna funkcija koja se poziva kada dodje do greske pri radu sa
   nitima. */
void pthread_error(char *message, int err)
{
  /* Funkcija strerror na osnovu vrednosti celobrojnog koda greske
     vraca pokazivac na string koji detaljnije opisuje gresku. */
  fprintf(stderr, "%s : %s\n", message, strerror(err));
  exit (EXIT_FAILURE);
}

/* Funkcija racuna n-ti clan Fibonacijevog niza. */
int fib (int n){
  if (n == 1 || n == 0)
    return 1;
  else {
    int fib_1 = fib(n-1);
    int fib_2 = fib(n-2);

    /* U slucaju bilo kog prekoracenja prekida se program. */
    if (fib_1 < 0){
      fprintf (stderr, "%d-ti clan Fibonacijevog niza prevelik za tip int.\n", n-1);
      exit (EXIT_FAILURE);
    }
    if (fib_2 < 0){
      fprintf (stderr, "%d-ti clan Fibonacijevog niza prevelik za tip int.\n", n-2);
      exit (EXIT_FAILURE);
    }
    return fib_1 + fib_2;
  }
}

/* Funkcija uvecava vrednost promenljive x NUM_LOOPS puta za po 1. */
void* fun(void *arg)
{
  int n = *((int*)arg);
  int fib_n = fib (*((int*)arg));
  if (fib_n >= 0)
    printf ("fib (%d) = %d\n", n, fib_n);
  else
    fprintf (stderr, "%d-ti clan Fibonacijevog niza prevelik za tip int.\n", n);
    
  return ((void*)0);
}

int main()
{
  int err;
  int i;
  pthread_t tid_a[NUM_THREADS];
  int *num1, *num2;

  /* Alociranje memorije za argumente niti i inicijalizacija vrednosti. */
  num1 = (int *)malloc (sizeof (int));
  num2 = (int *)malloc (sizeof (int));
  if (num1 == NULL || num2 == NULL)
    error_fatal ("malloc error");
  scanf ("%d%d", num1, num2);

  /* Kreiranje niti. */
  if( (err = pthread_create(&tid_a[0], NULL, fun, num1)) != 0)
    pthread_error ("thread create error", err);
  if( (err = pthread_create(&tid_a[1], NULL, fun, num2)) != 0)
    pthread_error ("thread create error", err);

  /* Cekanje na sve niti. */
  for(i = 0; i < NUM_THREADS; i++)
    if( (err = pthread_join(tid_a[i], NULL)) != 0)
      pthread_error ("thread join error", err);

  /* Dealokacija memorije. */
  free (num1);
  free (num2);

  return 0;
}
