#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <assert.h>
#include <pthread.h>

/* Maksimalna dimenzija matrice. */
#define MAX 10

/* Deklaracije matrica i njihove dimenzije. */
int a[MAX][MAX];
int b[MAX][MAX];
int c[MAX][MAX];
int n;

typedef struct _polje {
  int vrsta;
  int kolona;
} polje;

void error_fatal (char *message){
  perror (message);
  exit (EXIT_FAILURE);
}

void pthread_error (char *message, int err){
  fprintf (stderr, "%s : %s\n", message, strerror (err));
  exit (EXIT_FAILURE);
}

/* Funkcija ucitava dimenziju i dve matrice. */
void ucitaj_matrice (char *file){
  FILE *f;
  int i, j;

  if ( (f = fopen (file, "r")) == NULL)
    error_fatal ("read matrices error");

  fscanf (f, "%d", &n);

  for (i=0; i<n; i++)
    for (j=0; j<n; j++)
      fscanf (f, "%d", &a[i][j]);

  for (i=0; i<n; i++)
    for (j=0; j<n; j++)
      fscanf (f, "%d", &b[i][j]);

  fclose (f);
}

/* Pomnozi i-tu vrstu i j-tu kolonu matrica. */
void* pomnozi_vrstu_i_kolonu (void *arg){
  int s = 0;
  int k;
  int i = ((polje*)arg)->vrsta;
  int j = ((polje*)arg)->kolona;
  
  /* Izracunavanje proizvoda vrste i kolone i smestanje u matricu proizvoda. */
  for (k=0; k<n; k++)
    s += a[i][k] * b[k][j];
  c[i][j] = s;

  /* Kontrolni ispis koji se po potrebi moze ukljuciti/iskljuciti. */
  //  printf ("(%d,%d) = %d\n", i, j, s); 
  
  /* Oslobadjanje dinamicki alocirane memorije. */
  free (arg);

  /* Povratak vrednosti koja oznacava da je sve u redu. */
  return ((void*)0);
}

int main(int argc, char **argv){
  int err;
  int i, j;
  polje *p;
  pthread_t* tid_a;

  if (argc < 2){
    fprintf (stderr, "Potrebno ime datoteke sa matricama kao argument programa.\n");
    exit (EXIT_FAILURE);
  }
  ucitaj_matrice (argv[1]);

  /* Alociranje niza za identifikatore niti. */
  tid_a = malloc(n*n * sizeof(pthread_t));
  if(tid_a == NULL)
    error_fatal ("malloc error");

  for (i=0; i<n; i++)
    for (j=0; j<n; j++){
      /* Svaka pozicija koja se mnozi mora imati svoju strukturu!
	 Pogledati sta se desava kada se malloc prebaci pre petlji a
	 ukloni free iz pomnozi_vrstu_i_kolonu! (Funkcija se izvrsava
	 za nepredvidljive vrednosti indeksa). */
      p = malloc (sizeof (polje));
      if (p == NULL)
	error_fatal ("malloc error");

      p->vrsta = i;
      p->kolona = j;
      if ( (err = pthread_create (&tid_a[i*n+j], NULL, pomnozi_vrstu_i_kolonu, p)) != 0)
	pthread_error ("thread create error", err);
    }

  /* Cekanje na sve niti. */
  for(i = 0; i < n*n; i++){
    if( (err = pthread_join(tid_a[i], NULL)) != 0)
      pthread_error ("thread join error", err);
  }

  /* Stampa matrice proizvoda. */
  for (i=0; i<n; i++){
    for (j=0; j<n; j++)
      printf ("%-5d ", c[i][j]);
    printf ("\n");
  }

  /* Oslobadjanje memorije alocirane za tid_a. */
  free (tid_a);
  return 0;
}

  
      

